"""Builderers file config is a thin wrapper around this library."""

import collections
import dataclasses
import os
import posixpath
import subprocess
import typing
import uuid


@dataclasses.dataclass(frozen=True)
class Action:
    name: str
    commands: list[list[str]]


class Builderer:
    """The Builderer class is used to collect build tasks and issue them afterwards."""

    def __init__(
        self,
        *,
        registry: str | None = None,
        prefix: str | None = None,
        push: bool = True,
        cache: bool = False,
        verbose: bool = False,
        tags: list[str] = ["latest"],
        simulate: bool = False,
        backend: typing.Literal["docker", "podman"] = "docker",
    ) -> None:
        """Builderer runs commands inside in two queues. A action queue and a post queue.
        First the action queue gets handled (FIFO) then the corresponding post actions get called in reversed order (LIFO)
        Pushing is done as a post steps. This means a build is only pushed if builder of all images was successfull.

        Args:
            registry (str | None, optional): Registry URL. Defaults to None.
            prefix (str | None, optional): Registry folder / namespace / user. Defaults to None.
            push (bool, optional): Whether to allow pushing images. Defaults to True.
            cache (bool, optional): Allow using cached images. Defaults to False.
            verbose (bool, optional): Verbose output. Defaults to False.
            tags (list[str], optional): Tags to use. Defaults to ["latest"].
            simulate (bool, optional): Prevent issuing commands. Defaults to False.
            backend (typing.Literal["docker", "podman"], optional): Overwrite backend to use. Defaults to "docker".
        """
        self.tags = tags
        self.registry = registry
        self.prefix = prefix
        self.cache = cache
        self.backend = backend
        self.simulate = simulate
        self.verbose = verbose
        self.push = push

        self._actions: collections.deque[Action] = collections.deque()
        self._post: collections.deque[Action] = collections.deque()

    def action(self, name: str, commands: list[list[str]], post: bool) -> None:
        """A generic action with multiple commands.

        Hint: Use this mechanism if other commands aren't sufficient for your usecase.

        Args:
            name (str): Name of the action
            commands (list[list[str]]): List of commands. Each command is a list of strings: the executable followed by arguments.
            post (bool): Whether to add the action to the post queue.
        """
        item = Action(name=name, commands=commands)

        if post:
            self._post.appendleft(item)
        else:
            self._actions.append(item)

    def _full_image_name(self, name: str) -> str:
        return posixpath.join(self.registry or "", self.prefix or "", name)

    def _build_cmd(self, full_name: str) -> list[str]:
        tags = [i for tag in self.tags for i in ["-t", f"{full_name}:{tag}"]]
        cache = ["--no-cache"] if not self.cache else []

        return [self.backend, "build", *tags, *cache]

    def _run_cmd(self, command: list[str]) -> tuple[int, bytes]:
        if self.simulate:
            return 0, b""

        if self.verbose:
            proc = subprocess.run(command)
        else:
            proc = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        return proc.returncode, proc.stdout or b""

    def build_image(
        self,
        directory: str,
        *,
        dockerfile: str | None = None,
        name: str | None = None,
        push: bool = True,
        qualified: bool = True,
    ) -> None:
        """Build a docker image and push it to the registry.

        Args:
            directory (str): Directory containing the build context.
            dockerfile (str | None, optional): Path to Dockerfile. Name of the resulting image. Defaults to <directory>/Dockerfile.
            name (str | None, optional): Name of the resulting image. Defaults to the name of the Dockerfiles parent directory.
            push (bool, optional): Whether to push the image. Defaults to True.
            qualified (bool, optional): Whether to add the registry path and prefix to the image name. Defaults to True.
        """
        if dockerfile is None:
            dockerfile = os.path.join(directory, "Dockerfile")

        if name is None:
            name = os.path.basename(directory)

        image_name = self._full_image_name(name) if qualified else name

        self.action(
            name=f"Building image: {name}",
            commands=[[*self._build_cmd(image_name), "-f", dockerfile, directory]],
            post=False,
        )

        if not push or not self.push:
            return

        self.action(
            name=f"Pushing image: {name}",
            commands=[[self.backend, "push", f"{image_name}:{tag}"] for tag in self.tags],
            post=True,
        )

    def extract_from_image(self, image: str, path: str, *dest: str) -> None:
        """Copy a file from within a docker image.

        Args:
            image (str): Name of the image to copy from.
            path (str): Source path inside the image.
            dest (str): Destination paths. The file will be copied to all destinations individually.
        """
        image_name = self._full_image_name(image)
        container_name = str(uuid.uuid4())

        self.action(
            name=f"Extracting from image: {path} -> {', '.join(dest)}",
            commands=[
                [self.backend, "container", "create", "--name", container_name, image_name],
                *[[self.backend, "container", "cp", f"{container_name}:{path}", dst] for dst in dest],
                [self.backend, "container", "rm", "-f", container_name],
            ],
            post=False,
        )

    def forward_image(self, name: str, *, new_name: str | None = None) -> None:
        """Pulls an image from a registry, retags it and pushes it using the new names.

        Args:
            name (str): image name to pull
            new_name (str | None, optional): Set a new name for the image. By default the basename of the pulled image without the tag is used. Defaults to None.
        """
        if new_name is None:
            new_name = os.path.basename(name).split(":")[0]

        image_name = self._full_image_name(new_name)

        self.action(
            name=f"Forwarding image: {name} -> {new_name}",
            commands=[
                [self.backend, "pull", name],
                *[[self.backend, "tag", name, f"{image_name}:{tag}"] for tag in self.tags],
            ],
            post=False,
        )

        if not self.push:
            return

        self.action(
            name=f"Pushing image: {new_name}",
            commands=[[self.backend, "push", f"{image_name}:{tag}"] for tag in self.tags],
            post=True,
        )

    def pull_image(self, name: str) -> None:
        """Pulls an image from a registry. This might be usefull to ensure a local image is up to date (e.g. for local builds)

        Args:
            name (str): image name to pull.
        """
        self.action(
            name=f"Pulling image: {name}",
            commands=[[self.backend, "pull", name]],
            post=False,
        )

    def run(self) -> int:
        """After adding all steps. This method will start to issue all commands. It stops when done or a command fails.

        Returns:
            int: return code. On success this will be zero. Otherwise it will be the return code of the failed command.
        """
        for queue in [self._actions, self._post]:
            for action in queue:
                print(action.name, flush=True)

                for command in action.commands:
                    if self.verbose:
                        print(f"{command}", flush=True)

                    returncode, stdout = self._run_cmd(command)

                    if returncode != 0:
                        print("Encountered error running:", flush=True)

                        if not self.verbose:
                            print(stdout.decode(), flush=True)

                        return returncode

        return 0



import typing
from functools import partial

import pydantic
import pytest

import builderer.config


class AnyCaller:
    def __init__(self) -> None:
        self.calls: list[tuple[typing.Any, ...]] = []

    def _add_call(self, func_name: str, *args: typing.Any, **kwargs: typing.Any) -> None:
        self.calls.append((func_name, args, kwargs))

    def __getattr__(self, attr: str) -> typing.Any:
        if attr not in self.__dict__:
            return partial(self._add_call, attr)
        return super().__getattr__(attr)  # type: ignore


@pytest.fixture
def dummy_builderer() -> AnyCaller:
    return AnyCaller()


def test_action(dummy_builderer: AnyCaller) -> None:
    tester = builderer.config.Action(
        type="action",
        name="example",
        commands=[["command"], ["second", "--command", "foo"]],
        post=True,
    )
    tester.add_to(dummy_builderer)  # type: ignore

    assert dummy_builderer.calls == [
        (
            "action",
            tuple(),
            {"name": "example", "commands": [["command"], ["second", "--command", "foo"]], "post": True},
        )
    ]


def test_build_image(dummy_builderer: AnyCaller) -> None:
    tester = builderer.config.BuildImage(
        type="build_image",
        directory="some folder",
        dockerfile="some/docker/file",
        name="custom-name",
        push=False,
        qualified=False,
    )
    tester.add_to(dummy_builderer)  # type: ignore

    assert dummy_builderer.calls == [
        (
            "build_image",
            tuple(),
            {
                "directory": "some folder",
                "dockerfile": "some/docker/file",
                "name": "custom-name",
                "push": False,
                "qualified": False,
            },
        )
    ]


def test_build_images(dummy_builderer: AnyCaller) -> None:
    tester = builderer.config.BuildImages(
        type="build_images",
        directories=["folder1", "folder2"],
        push=False,
        qualified=False,
    )
    tester.add_to(dummy_builderer)  # type: ignore

    assert dummy_builderer.calls == [
        (
            "build_image",
            tuple(),
            {
                "directory": "folder1",
                "push": False,
                "qualified": False,
            },
        ),
        (
            "build_image",
            tuple(),
            {
                "directory": "folder2",
                "push": False,
                "qualified": False,
            },
        ),
    ]


def test_extract_from_image(dummy_builderer: AnyCaller) -> None:
    tester = builderer.config.ExtractFromImage(
        type="extract_from_image",
        image="some-image",
        path="/path/to/file",
        dest=["a", "b/c", "something"],
    )
    tester.add_to(dummy_builderer)  # type: ignore

    assert dummy_builderer.calls == [
        (
            "extract_from_image",
            ("some-image", "/path/to/file", "a", "b/c", "something"),
            {},
        )
    ]


def test_forward_image(dummy_builderer: AnyCaller) -> None:
    tester = builderer.config.ForwardImage(
        type="forward_image",
        name="image-name",
        new_name="something-else",
    )
    tester.add_to(dummy_builderer)  # type: ignore

    assert dummy_builderer.calls == [
        (
            "forward_image",
            tuple(),
            {
                "name": "image-name",
                "new_name": "something-else",
            },
        )
    ]


def test_pull_image(dummy_builderer: AnyCaller) -> None:
    tester = builderer.config.PullImage(type="pull_image", name="some-image-name")
    tester.add_to(dummy_builderer)  # type: ignore

    assert dummy_builderer.calls == [("pull_image", tuple(), {"name": "some-image-name"})]


def test_pull_images(dummy_builderer: AnyCaller) -> None:
    tester = builderer.config.PullImages(type="pull_images", names=["some-image-name", "second"])
    tester.add_to(dummy_builderer)  # type: ignore

    assert dummy_builderer.calls == [
        ("pull_image", tuple(), {"name": "some-image-name"}),
        ("pull_image", tuple(), {"name": "second"}),
    ]


@pytest.mark.parametrize(
    ("data", "error", "error_texts"),
    [
        ({}, pydantic.ValidationError, ["value_error.missing", "'loc': ('steps'"]),
        ({"steps": [{}]}, ValueError, ["'value_error'", "malformed step: 'type' is required!"]),
        ({"steps": [1]}, pydantic.ValidationError, ["'type_error.dict'", "value is not a valid dict"]),
        ({"steps": [{"type": "unknown"}]}, ValueError, ["'value_error'", "Unknown step type unknown"]),
    ],
)
def test_builderer_config_errors(data: typing.Any, error: typing.Type[Exception], error_texts: list[str]) -> None:
    with pytest.raises(error) as e:
        builderer.config.BuildererConfig.parse_obj(data)

    for text in error_texts:
        assert text in str(e)



# arguments in parameters and cli
arg_registry_title = "Registry URL"
arg_registry_desc = "Set the registry url. You may include a port using the colon notation (example.com:3000/). This is needed when using a non standard port. Unset by default."
arg_prefix_title = "Registry directory"
arg_prefix_desc = "Set the directory for all images. This is the image component between registry url and image name. For example on docker hub this is used for the username. Unset by default."
arg_push_title = "Allow pushing images"
arg_push_desc = "Whether to allow pushing images."
arg_cache_title = "Allow caching images"
arg_cache_desc = "Whether to allow using cached images. This is especially usefull for local builds."
arg_verbose_title = "Verbose output"
arg_verbose_desc = "Show issued commands and their live output."
arg_tags_title = "Tags to use"
arg_tags_desc = "One or multiple tags to use for each image. Default to ['latest']"
arg_simulate_title = "Simulate execution"
arg_simulate_desc = "Prevent issuing any commands just do the printing."
arg_backend_title = "Build Backend"
arg_backend_desc = "Overwrite the backend used to build, tag and pull images. Defaults to 'docker'"

# just cli
arg_cli_config = "Path to %(prog)s yaml configuration file. Defaults to '.builderer.yml'"
arg_cli_no_push = "Prevent pushing images in all steps."

# just parameters
conf_parameters = "Overwrite default parameters. Values set here will be overwritten by command line arguments."
conf_steps = "List of steps to execute."

# steps
step_type = "Type of the step"

step_action_name = "Name printed before running the action"
step_action_commands = "List of commands. Each command is a list of strings: the executable followed by arguments."
step_action_post = "Whether to add the action to the post queue"

step_build_directory = "Directory containing the Dockerfile. This is also used as the build context."
step_build_directories = "Directories containing each containing Dockerfile."
step_build_dockerfile = "Path to Dockerfile. Name of the resulting image. Defaults to <directory>/Dockerfile."
step_build_name = "Name of the resulting image. Defaults to the name of the Dockerfiles parent directory."
step_build_push = "Whether to push the image. Defaults to True."
step_build_qualified = "Whether to add the registry path and prefix to the image name. Defaults to True."

step_extract_image = "Name of the image to copy from."
step_extract_path = "Source path inside the image."
step_extract_dest = "Destination paths. The file will be copied to all destinations individually."

step_forward_name = "Image name to forward."
step_forward_new_name = (
    "Set a new name for the image. By default the basename of the pulled image without the tag is used."
)

step_pull_name = "Image name to pull."
step_pull_names = "Image names to pull."



import pathlib
import typing

import pydantic
import yaml

import builderer._documentation as docs
from builderer import builderer


class _BaseModel(pydantic.BaseModel):
    class Config:
        extra = pydantic.Extra.forbid


class Action(_BaseModel):
    type: typing.Literal["action"] = pydantic.Field(description=docs.step_type)
    name: str = pydantic.Field(description=docs.step_action_name)
    commands: list[list[str]] = pydantic.Field(description=docs.step_action_commands)
    post: bool = pydantic.Field(description=docs.step_action_post)

    def add_to(self, builderer: builderer.Builderer) -> None:
        builderer.action(name=self.name, commands=self.commands, post=self.post)


class BuildImage(_BaseModel):
    type: typing.Literal["build_image"] = pydantic.Field(description=docs.step_type)
    directory: str = pydantic.Field(description=docs.step_build_directory)
    dockerfile: str | None = pydantic.Field(default=None, description=docs.step_build_dockerfile)
    name: str | None = pydantic.Field(default=None, description=docs.step_build_name)
    push: bool = pydantic.Field(default=True, description=docs.step_build_push)
    qualified: bool = pydantic.Field(default=True, description=docs.step_build_qualified)

    def add_to(self, builderer: builderer.Builderer) -> None:
        builderer.build_image(
            directory=self.directory,
            dockerfile=self.dockerfile,
            name=self.name,
            push=self.push,
            qualified=self.qualified,
        )


class BuildImages(_BaseModel):
    type: typing.Literal["build_images"] = pydantic.Field(description=docs.step_type)
    directories: list[str] = pydantic.Field(description=docs.step_build_directories)
    push: bool = pydantic.Field(default=True, description=docs.step_build_push)
    qualified: bool = pydantic.Field(default=True, description=docs.step_build_qualified)

    def add_to(self, builderer: builderer.Builderer) -> None:
        for directory in self.directories:
            builderer.build_image(directory=directory, push=self.push, qualified=self.qualified)


class ExtractFromImage(_BaseModel):
    type: typing.Literal["extract_from_image"] = pydantic.Field(description=docs.step_type)
    image: str = pydantic.Field(description=docs.step_extract_image)
    path: str = pydantic.Field(description=docs.step_extract_path)
    dest: list[str] = pydantic.Field(description=docs.step_extract_dest)

    def add_to(self, builderer: builderer.Builderer) -> None:
        builderer.extract_from_image(self.image, self.path, *self.dest)


class ForwardImage(_BaseModel):
    type: typing.Literal["forward_image"] = pydantic.Field(description=docs.step_type)
    name: str = pydantic.Field(description=docs.step_forward_name)
    new_name: str | None = pydantic.Field(default=None, description=docs.step_forward_new_name)

    def add_to(self, builderer: builderer.Builderer) -> None:
        builderer.forward_image(name=self.name, new_name=self.new_name)


class PullImage(_BaseModel):
    type: typing.Literal["pull_image"] = pydantic.Field(description=docs.step_type)
    name: str = pydantic.Field(description=docs.step_pull_name)

    def add_to(self, builderer: builderer.Builderer) -> None:
        builderer.pull_image(name=self.name)


class PullImages(_BaseModel):
    type: typing.Literal["pull_images"] = pydantic.Field(description=docs.step_type)
    names: list[str] = pydantic.Field(description=docs.step_pull_names)

    def add_to(self, builderer: builderer.Builderer) -> None:
        for name in self.names:
            builderer.pull_image(name=name)


class Parameters(_BaseModel):
    registry: str | None = pydantic.Field(None, title=docs.arg_registry_title, description=docs.arg_registry_desc)
    prefix: str | None = pydantic.Field(None, title=docs.arg_prefix_title, description=docs.arg_prefix_desc)
    push: bool | None = pydantic.Field(None, title=docs.arg_push_title, description=docs.arg_push_desc)
    cache: bool | None = pydantic.Field(None, title=docs.arg_cache_title, description=docs.arg_cache_desc)
    verbose: bool | None = pydantic.Field(None, title=docs.arg_verbose_title, description=docs.arg_verbose_desc)
    tags: list[str] | None = pydantic.Field(None, title=docs.arg_tags_title, description=docs.arg_tags_desc)
    simulate: bool | None = pydantic.Field(None, title=docs.arg_simulate_title, description=docs.arg_simulate_desc)
    backend: typing.Literal["docker", "podman"] | None = pydantic.Field(
        None, title=docs.arg_backend_title, description=docs.arg_backend_desc
    )


class BuildererConfig(_BaseModel):
    steps: list[
        Action | BuildImage | BuildImages | ExtractFromImage | ForwardImage | PullImage | PullImages
    ] = pydantic.Field(description=docs.conf_steps)

    parameters: Parameters = pydantic.Field(
        default_factory=Parameters,  # pyright: ignore
        description=docs.conf_parameters,
    )

    @pydantic.validator("steps", pre=True, each_item=True)
    def parse_steps_by_type(cls, v: typing.Any) -> typing.Any:
        if not isinstance(v, dict):
            return v

        if "type" not in v:
            raise ValueError("malformed step: 'type' is required!")

        class_name = "".join(x.title() for x in v["type"].split("_"))

        for step_type in cls.__fields__["steps"].type_.__args__:
            if class_name == step_type.__name__:
                return step_type.parse_obj(v)

        raise ValueError(f"Unknown step type {v['type']}")

    @staticmethod
    def load(path: str | pathlib.Path) -> "BuildererConfig":
        with open(path, "rt") as f:
            return BuildererConfig.parse_obj(yaml.safe_load(f))



import argparse
from typing import Any

import pydantic

import builderer._documentation as docs
from builderer import __version__
from builderer.builderer import Builderer
from builderer.config import BuildererConfig


def parse_args(argv: list[str] | None = None) -> tuple[str, dict[str, Any]]:
    parser = argparse.ArgumentParser(
        prog="builderer",
        description="Building and pushing containers. \n\nCommand line arguments take precedence over file configuration which in turn takes precedence over default values",
        epilog="This program is intended to run locally and inside ci/cd jobs.",
    )

    parser.add_argument("--registry", type=str, default=None, help=docs.arg_registry_desc)
    parser.add_argument("--prefix", type=str, default=None, help=docs.arg_prefix_desc)
    parser.add_argument("--tags", nargs="+", type=str, default=None, help=docs.arg_tags_desc)
    parser.add_argument("--no-push", action="store_false", dest="push", default=None, help=docs.arg_cli_config)
    parser.add_argument("--cache", action="store_true", default=None, help=docs.arg_cache_desc)
    parser.add_argument("--verbose", action="store_true", default=None, help=docs.arg_verbose_desc)
    parser.add_argument("--simulate", action="store_true", default=None, help=docs.arg_simulate_desc)
    parser.add_argument("--backend", choices=["docker", "podman"], help=docs.arg_backend_desc)
    parser.add_argument("--config", type=str, default=".builderer.yml", help=docs.arg_cli_config)
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    arguments = parser.parse_args(argv)

    return arguments.config, {k: v for k, v in vars(arguments).items() if v is not None and k != "config"}


def main(argv: list[str] | None = None) -> int:
    config_path, cli_args = parse_args(argv)

    try:
        config = BuildererConfig.load(config_path)
    except (FileNotFoundError, pydantic.ValidationError) as e:
        print(e)
        return 1

    builderer_args = config.parameters.dict(exclude_none=True) | cli_args

    builderer = Builderer(**builderer_args)

    for step in config.steps:
        step.add_to(builderer)

    return builderer.run()


if __name__ == "__main__":
    raise SystemExit(main())  # pragma: no cover



import pathlib

import builderer.config


def test_load_minimal(datadir: pathlib.Path) -> None:
    file_config = builderer.config.BuildererConfig.load(datadir / "minimal.yml")

    assert file_config.dict() == {
        "steps": [],
        "parameters": {
            "registry": None,
            "prefix": None,
            "push": None,
            "cache": None,
            "verbose": None,
            "tags": None,
            "simulate": None,
            "backend": None,
        },
    }


def test_load_example(datadir: pathlib.Path) -> None:
    file_config = builderer.config.BuildererConfig.load(datadir / "example.yml")

    assert file_config.dict() == {
        "parameters": {
            "registry": "registry.example.com:12345",
            "prefix": "username",
            "push": False,
            "cache": False,
            "verbose": False,
            "tags": ["a", "b"],
            "simulate": True,
            "backend": "podman",
        },
        "steps": [],
    }


def test_load_example_workspace(datadir: pathlib.Path) -> None:
    file_config = builderer.config.BuildererConfig.load(datadir / "example_workspace" / ".builderer.yml")

    assert file_config.dict() == {
        "parameters": {
            "registry": "registry.example.com",
            "prefix": "foo",
            "push": None,
            "cache": None,
            "verbose": None,
            "tags": None,
            "simulate": None,
            "backend": None,
        },
        "steps": [
            {"type": "pull_images", "names": ["docker.io/python:alpine", "docker.io/nginx:alpine"]},
            {"type": "forward_image", "name": "docker.io/redis:alpine", "new_name": None},
            {"type": "build_images", "directories": ["frontend", "backend"], "push": True, "qualified": True},
        ],
    }



import pathlib
import re
import subprocess
import typing
from unittest.mock import call

import pytest
from pytest_mock import MockerFixture

import builderer.__main__


@pytest.mark.parametrize(
    ("input_args", "expected_config", "expected_args"),
    [
        (
            [],
            ".builderer.yml",
            {},
        ),
        (
            [
                "--registry",
                "reg.examle.com:6789",
                "--prefix",
                "user",
                "--tags",
                "foo",
                "bar",
                "baz",
                "--no-push",
                "--cache",
                "--verbose",
                "--simulate",
                "--backend",
                "podman",
                "--config",
                "test.yaml",
            ],
            "test.yaml",
            {
                "registry": "reg.examle.com:6789",
                "prefix": "user",
                "tags": ["foo", "bar", "baz"],
                "push": False,
                "cache": True,
                "verbose": True,
                "simulate": True,
                "backend": "podman",
            },
        ),
    ],
)
def test_parse_args(input_args: list[str], expected_config: str, expected_args: dict[str, typing.Any]) -> None:
    actual_config, actual_args = builderer.__main__.parse_args(input_args)
    assert actual_config == expected_config
    assert actual_args == expected_args


@pytest.mark.parametrize(
    ("flag", "pattern"),
    [
        ("--help", "^usage: builderer.*options"),
        ("-h", "^usage: builderer.*options"),
        ("--version", r"^builderer \d+\.\d+\.\d+$"),
    ],
)
def test_parse_args_special_action(flag: str, pattern: str, capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit, match="^0$"):
        builderer.__main__.parse_args([flag])

    captured = capsys.readouterr()

    assert captured.err == ""
    assert re.match(pattern, captured.out, re.DOTALL) is not None


@pytest.mark.parametrize("verbose", [True, False])
def test_main_simulate_example_workspace(
    datadir: pathlib.Path, capsys: pytest.CaptureFixture[str], verbose: bool
) -> None:
    run_config = datadir / "example_workspace" / ".builderer.yml"
    arguments = ["--config", str(run_config), "--simulate"] + (["--verbose"] if verbose else [])

    return_code = builderer.__main__.main(arguments)
    captured = capsys.readouterr()

    assert return_code == 0
    assert captured.err == ""

    if verbose:
        assert captured.out.split("\n") == [
            "Pulling image: docker.io/python:alpine",
            "['docker', 'pull', 'docker.io/python:alpine']",
            "Pulling image: docker.io/nginx:alpine",
            "['docker', 'pull', 'docker.io/nginx:alpine']",
            "Forwarding image: docker.io/redis:alpine -> redis",
            "['docker', 'pull', 'docker.io/redis:alpine']",
            "['docker', 'tag', 'docker.io/redis:alpine', 'registry.example.com/foo/redis:latest']",
            "Building image: frontend",
            "['docker', 'build', '-t', 'registry.example.com/foo/frontend:latest', '--no-cache', '-f', 'frontend/Dockerfile', 'frontend']",
            "Building image: backend",
            "['docker', 'build', '-t', 'registry.example.com/foo/backend:latest', '--no-cache', '-f', 'backend/Dockerfile', 'backend']",
            "Pushing image: backend",
            "['docker', 'push', 'registry.example.com/foo/backend:latest']",
            "Pushing image: frontend",
            "['docker', 'push', 'registry.example.com/foo/frontend:latest']",
            "Pushing image: redis",
            "['docker', 'push', 'registry.example.com/foo/redis:latest']",
            "",
        ]

    else:
        assert captured.out.split("\n") == [
            "Pulling image: docker.io/python:alpine",
            "Pulling image: docker.io/nginx:alpine",
            "Forwarding image: docker.io/redis:alpine -> redis",
            "Building image: frontend",
            "Building image: backend",
            "Pushing image: backend",
            "Pushing image: frontend",
            "Pushing image: redis",
            "",
        ]


def test_main_run_example_workspace(
    datadir: pathlib.Path, capsys: pytest.CaptureFixture[str], mocker: MockerFixture
) -> None:
    run_patch = mocker.patch("subprocess.run", return_value=subprocess.CompletedProcess([], 0))

    arguments = ["--config", str(datadir / "example_workspace" / ".builderer.yml"), "--verbose"]
    return_code = builderer.__main__.main(arguments)
    captured = capsys.readouterr()

    assert return_code == 0
    assert captured.err == ""

    assert captured.out.split("\n") == [
        "Pulling image: docker.io/python:alpine",
        "['docker', 'pull', 'docker.io/python:alpine']",
        "Pulling image: docker.io/nginx:alpine",
        "['docker', 'pull', 'docker.io/nginx:alpine']",
        "Forwarding image: docker.io/redis:alpine -> redis",
        "['docker', 'pull', 'docker.io/redis:alpine']",
        "['docker', 'tag', 'docker.io/redis:alpine', 'registry.example.com/foo/redis:latest']",
        "Building image: frontend",
        "['docker', 'build', '-t', 'registry.example.com/foo/frontend:latest', '--no-cache', '-f', 'frontend/Dockerfile', 'frontend']",
        "Building image: backend",
        "['docker', 'build', '-t', 'registry.example.com/foo/backend:latest', '--no-cache', '-f', 'backend/Dockerfile', 'backend']",
        "Pushing image: backend",
        "['docker', 'push', 'registry.example.com/foo/backend:latest']",
        "Pushing image: frontend",
        "['docker', 'push', 'registry.example.com/foo/frontend:latest']",
        "Pushing image: redis",
        "['docker', 'push', 'registry.example.com/foo/redis:latest']",
        "",
    ]

    expected_calls = [
        call(["docker", "pull", "docker.io/python:alpine"]),
        call(["docker", "pull", "docker.io/nginx:alpine"]),
        call(["docker", "pull", "docker.io/redis:alpine"]),
        call(["docker", "tag", "docker.io/redis:alpine", "registry.example.com/foo/redis:latest"]),
        call(
            [
                "docker",
                "build",
                "-t",
                "registry.example.com/foo/frontend:latest",
                "--no-cache",
                "-f",
                "frontend/Dockerfile",
                "frontend",
            ]
        ),
        call(
            [
                "docker",
                "build",
                "-t",
                "registry.example.com/foo/backend:latest",
                "--no-cache",
                "-f",
                "backend/Dockerfile",
                "backend",
            ]
        ),
        call(["docker", "push", "registry.example.com/foo/backend:latest"]),
        call(["docker", "push", "registry.example.com/foo/frontend:latest"]),
        call(["docker", "push", "registry.example.com/foo/redis:latest"]),
    ]

    assert run_patch.call_count == len(expected_calls)
    run_patch.assert_has_calls(expected_calls, any_order=False)


def test_main_run_unknown_workspace(tmp_path: pathlib.Path, capsys: pytest.CaptureFixture[str]) -> None:
    arguments = ["--config", str(tmp_path / ".builderer.yml")]

    return_code = builderer.__main__.main(arguments)
    captured = capsys.readouterr()

    assert return_code == 1
    assert captured.err == ""
    assert "No such file or directory:" in captured.out



import pathlib
import sys

import pytest
from pytest_mock import MockerFixture

from builderer.builderer import Builderer


@pytest.fixture
def empty_builderer() -> Builderer:
    return Builderer()


def test_empty_builderer(empty_builderer: Builderer) -> None:
    assert empty_builderer.tags == ["latest"]
    assert empty_builderer.registry is None
    assert empty_builderer.prefix is None
    assert empty_builderer.cache is False
    assert empty_builderer.backend == "docker"
    assert empty_builderer.simulate is False
    assert empty_builderer.verbose is False
    assert empty_builderer.push is True
    assert len(empty_builderer._actions) == 0
    assert len(empty_builderer._post) == 0


@pytest.fixture
def sim_builderer(empty_builderer: Builderer) -> Builderer:
    empty_builderer.simulate = True
    empty_builderer.verbose = True
    return empty_builderer


@pytest.mark.parametrize("name", ["some_name", "nothing", "foo"])
@pytest.mark.parametrize("registry", ["localhost:3000/", "registry.example.com/", None])
@pytest.mark.parametrize("push", [True, False])
def test_build_image(
    sim_builderer: Builderer, capsys: pytest.CaptureFixture[str], name: str, registry: str | None, push: bool
) -> None:
    sim_builderer.registry = registry
    sim_builderer.build_image(name, push=push)
    ret = sim_builderer.run()
    captured = capsys.readouterr()

    prefix = registry or ""

    push_output = [f"Pushing image: {name}", f"['docker', 'push', '{prefix}{name}:latest']"] if push else []

    assert ret == 0
    assert captured.err == ""
    assert captured.out.split("\n") == [
        f"Building image: {name}",
        f"['docker', 'build', '-t', '{prefix}{name}:latest', '--no-cache', '-f', '{name}/Dockerfile', '{name}']",
        *push_output,
        "",
    ]


def test_extract_from_image(
    sim_builderer: Builderer, capsys: pytest.CaptureFixture[str], mocker: MockerFixture
) -> None:
    mocker.patch("uuid.uuid4", return_value="50f17813-ac32-4e69-9029-ea3f8e656125")

    sim_builderer.extract_from_image("localhost:3000/reg/some-image:42", "/build/result", "./a/", "b", "/opt/data")
    ret = sim_builderer.run()
    captured = capsys.readouterr()

    assert ret == 0
    assert captured.err == ""
    assert captured.out.split("\n") == [
        "Extracting from image: /build/result -> ./a/, b, /opt/data",
        "['docker', 'container', 'create', '--name', '50f17813-ac32-4e69-9029-ea3f8e656125', 'localhost:3000/reg/some-image:42']",
        "['docker', 'container', 'cp', '50f17813-ac32-4e69-9029-ea3f8e656125:/build/result', './a/']",
        "['docker', 'container', 'cp', '50f17813-ac32-4e69-9029-ea3f8e656125:/build/result', 'b']",
        "['docker', 'container', 'cp', '50f17813-ac32-4e69-9029-ea3f8e656125:/build/result', '/opt/data']",
        "['docker', 'container', 'rm', '-f', '50f17813-ac32-4e69-9029-ea3f8e656125']",
        "",
    ]


@pytest.mark.parametrize("push", [True, False])
def test_forward_image_simple(sim_builderer: Builderer, capsys: pytest.CaptureFixture[str], push: bool) -> None:
    sim_builderer.push = push
    sim_builderer.forward_image("registry.example.com:1234/foo/some-image:v123-test")
    ret = sim_builderer.run()
    captured = capsys.readouterr()

    push_output = (
        [
            "Pushing image: some-image",
            "['docker', 'push', 'some-image:latest']",
        ]
        if push
        else []
    )

    assert ret == 0
    assert captured.err == ""
    assert captured.out.split("\n") == [
        "Forwarding image: registry.example.com:1234/foo/some-image:v123-test -> some-image",
        "['docker', 'pull', 'registry.example.com:1234/foo/some-image:v123-test']",
        "['docker', 'tag', 'registry.example.com:1234/foo/some-image:v123-test', 'some-image:latest']",
        *push_output,
        "",
    ]


@pytest.mark.parametrize("push", [True, False])
def test_forward_image_complex(sim_builderer: Builderer, capsys: pytest.CaptureFixture[str], push: bool) -> None:
    sim_builderer.push = push
    sim_builderer.tags = ["tag1", "3.14", "v3-alpine"]
    sim_builderer.registry = "some-reg.server.org:9001"
    sim_builderer.prefix = "my-prefix"
    sim_builderer.forward_image("registry.example.com:3333/bar/remote-image:42", new_name="new-image-name")
    ret = sim_builderer.run()
    captured = capsys.readouterr()

    push_output = (
        [
            "Pushing image: new-image-name",
            "['docker', 'push', 'some-reg.server.org:9001/my-prefix/new-image-name:tag1']",
            "['docker', 'push', 'some-reg.server.org:9001/my-prefix/new-image-name:3.14']",
            "['docker', 'push', 'some-reg.server.org:9001/my-prefix/new-image-name:v3-alpine']",
        ]
        if push
        else []
    )
    assert ret == 0
    assert captured.err == ""
    assert captured.out.split("\n") == [
        "Forwarding image: registry.example.com:3333/bar/remote-image:42 -> new-image-name",
        "['docker', 'pull', 'registry.example.com:3333/bar/remote-image:42']",
        "['docker', 'tag', 'registry.example.com:3333/bar/remote-image:42', 'some-reg.server.org:9001/my-prefix/new-image-name:tag1']",
        "['docker', 'tag', 'registry.example.com:3333/bar/remote-image:42', 'some-reg.server.org:9001/my-prefix/new-image-name:3.14']",
        "['docker', 'tag', 'registry.example.com:3333/bar/remote-image:42', 'some-reg.server.org:9001/my-prefix/new-image-name:v3-alpine']",
        *push_output,
        "",
    ]


@pytest.mark.parametrize("name", ["some_name", "localhost:3000/image", "ghcr.io/foo/bar"])
def test_pull_image(sim_builderer: Builderer, capsys: pytest.CaptureFixture[str], name: str) -> None:
    sim_builderer.pull_image(name)
    ret = sim_builderer.run()
    captured = capsys.readouterr()

    assert ret == 0
    assert captured.err == ""
    assert captured.out.split("\n") == [
        f"Pulling image: {name}",
        f"['docker', 'pull', '{name}']",
        "",
    ]


def test_run_error(empty_builderer: Builderer, capsys: pytest.CaptureFixture[str], tmp_path: pathlib.Path) -> None:
    empty_builderer.action(
        "Failing action", [[sys.executable, "-c", "print('example output'); raise SystemExit(42)"]], post=False
    )
    ret = empty_builderer.run()
    captured = capsys.readouterr()

    assert ret == 42
    assert captured.err == ""
    assert captured.out.split("\n") == [
        "Failing action",
        "Encountered error running:",
        "example output",
        "",
        "",
    ]
